@@ -2,7 +2,7 @@ package RedisDB;
 
 use strict;
 use warnings;
-our $VERSION = "2.08";
+our $VERSION = "2.10";
 $VERSION = eval $VERSION;
 
 use RedisDB::Error;
@@ -30,9 +30,10 @@ RedisDB - Perl extension to access redis database
 
 =head1 DESCRIPTION
 
-This module provides interface to access redis key-value store. It
-transparently handles disconnects and forks. It supports transactions,
-pipelining, and subscription mode.
+This module provides interface to access redis key-value store, it
+transparently handles disconnects and forks, supports transactions,
+pipelining, and subscription mode. Module includes XS and pure Perl
+versions of the parser.
 
 =head1 METHODS
 
@@ -69,12 +70,19 @@ connecting to the server. Database changes when you sending I<select> command
 to the server. You can get current database using I<selected_database> method.
 Default value is 0.
 
+=item raise_error
+
+By default if redis-server returned error reply, I<get_reply> method throws
+an exception of L<RedisDB::Error> type, if you set this parameter to false it
+will return error object instead.
+
 =item timeout
 
 IO timeout. With this option set, if IO operation has taken more than specified
-number of seconds, module will croak. Note, that some OSes do not support
-SO_RCVTIMEO, and SO_SNDTIMEO socket options, in this case timeout will not
-work.
+number of seconds, module will croak or return L<RedisDB::Error::EAGAIN> error
+object depending on L</raise_error> setting. Note, that some OSes do not
+support SO_RCVTIMEO, and SO_SNDTIMEO socket options, in this case timeout will
+not work.
 
 =item utf8
 
@@ -88,6 +96,28 @@ by default I<new> establishes connection to the server. If this parameter is
 set, then connection will be established when you will send a command to the
 server.
 
+=item reconnect_attempts
+
+this parameter allows you to specify how many attempts to (re)connect to the
+server should be made before returning error. Default value is 1, set to -1 if
+module should try to reconnect indefinitely.
+
+=item reconnect_delay_max
+
+module makes a delay before each new attempt to connect. Delay increases with
+each new attempt. This parameter allows you to specify maximum delay between
+attempts to reconnect. Default value is 10.
+
+=item on_connect_error
+
+this allows you to specify callback that will be invoked in the case of failed
+connection attempt. First argument to callback is a reference to the RedisDB
+object, and second is the error description. You must not invoke any methods on
+the object, but you can change I<port> and I<host>, or I<path> attributes.
+After callback returned, module tries to establish connection again. Default
+callback confesses. This may be useful to switch to reserve server if primary
+is down.
+
 =back
 
 =cut
@@ -101,9 +131,13 @@ sub new {
     }
     $self->{port} ||= 6379;
     $self->{host} ||= 'localhost';
+    $self->{raise_error}    = 1 unless exists $self->{raise_error};
     $self->{_replies}       = [];
     $self->{_to_be_fetched} = 0;
-    $self->{database} ||= 0;
+    $self->{database}            ||= 0;
+    $self->{reconnect_attempts}  ||= 1;
+    $self->{reconnect_delay_max} ||= 10;
+    $self->{on_connect_error}    ||= \&_on_connect_error;
     $self->_init_parser;
     $self->_connect unless $self->{lazy};
     return $self;
@@ -116,9 +150,10 @@ sub _init_parser {
 
 =head2 $self->execute($command, @arguments)
 
-send a command to the server and return the result. It will throw the exception
-if the server returns an error. It may be more convenient to use instead of
-this method wrapper named after the corresponding redis command. E.g.:
+send a command to the server and return the result. It will throw an exception
+if the server returns an error or return L<RedisDB::Error> depending on
+L</raise_error> parameter. It may be more convenient to use instead of this
+method wrapper named after the corresponding redis command. E.g.:
 
     $redis->execute('set', key => 'value');
     # is the same as
@@ -142,6 +177,12 @@ sub execute {
     return $self->get_reply;
 }
 
+sub _on_connect_error {
+    my ( $self, $err ) = @_;
+    my $server = $self->{path} || ("$self->{host}:$self->{port}");
+    confess "Couldn't connect to the redis server at $server: $!";
+}
+
 # establish connection to the server.
 
 sub _connect {
@@ -154,19 +195,34 @@ sub _connect {
 
     $self->{_pid} = $$;
 
-    if ( $self->{path} ) {
-        $self->{_socket} = IO::Socket::UNIX->new(
-            Type => SOCK_STREAM,
-            Peer => $self->{path},
-        ) or croak "Can't connect to redis server socket at `$self->{path}': $!";
+    delete $self->{_socket};
+    my $error;
+    while ( not $self->{_socket} ) {
+        if ( $self->{path} ) {
+            $self->{_socket} = IO::Socket::UNIX->new(
+                Type => SOCK_STREAM,
+                Peer => $self->{path},
+            ) or $error = $!;
+        }
+        else {
+            my $attempts = $self->{reconnect_attempts};
+            my $delay;
+            while ( not $self->{_socket} and $attempts ) {
+                sleep $delay if $delay;
+                $self->{_socket} = IO::Socket::INET->new(
+                    PeerAddr => $self->{host},
+                    PeerPort => $self->{port},
+                    Proto    => 'tcp',
+                    ( $self->{timeout} ? ( Timeout => $self->{timeout} ) : () ),
+                ) or $error = $!;
+                $delay = $delay ? ( 1 + rand ) * $delay : 1;
+                $delay = $self->{reconnect_delay_max} if $delay > $self->{reconnect_delay_max};
+                $attempts--;
+            }
+        }
     }
-    else {
-        $self->{_socket} = IO::Socket::INET->new(
-            PeerAddr => $self->{host},
-            PeerPort => $self->{port},
-            Proto    => 'tcp',
-            ( $self->{timeout} ? ( Timeout => $self->{timeout} ) : () ),
-        ) or croak "Can't connect to redis server $self->{host}:$self->{port}: $!";
+    continue {
+        $self->{on_connect_error}->( $self, $error ) unless $self->{_socket};
     }
 
     if ( $self->{timeout} ) {
@@ -246,7 +302,7 @@ sub _recv_data_nb {
             # if there are some replies lost
             confess "Server closed connection. Some data was lost."
               if $self->{_parser}->callbacks
-                  or $self->{_in_multi};
+              or $self->{_in_multi};
 
             # clean disconnect, try to reconnect
             $self->{warnings} and warn "Disconnected, trying to reconnect";
@@ -269,7 +325,7 @@ sub _queue {
 =head2 $self->send_command($command[, @arguments][, \&callback])
 
 send a command to the server. Returns true if the command was successfully
-sent, or dies if an error occured. Note, that it does not return reply from
+sent, or dies if an error occurred. Note, that it does not return reply from
 the server, you should retrieve it using the I<get_reply> method, or if I<callback>
 has been specified, it will be invoked upon receiving the reply with two
 arguments: the RedisDB object, and the reply from the server.  If the server
@@ -424,7 +480,9 @@ sub mainloop {
 
 =head2 $self->get_reply
 
-receive reply from the server. Method croaks if the server returns an error.
+Receive and return reply from the server. If the server returned an error,
+method throws L<RedisDB::Error> exception or returns L<RedisDB::Error> object,
+depending on I<raise_error> parameter, see I<new>.
 
 =cut
 
@@ -433,13 +491,22 @@ sub get_reply {
 
     croak "We are not waiting for reply"
       unless @{ $self->{_replies} }
-          or $self->{_to_be_fetched}
-          or $self->{_subscription_loop};
+      or $self->{_to_be_fetched}
+      or $self->{_subscription_loop};
     croak "You can't read reply in child process" unless $self->{_pid} == $$;
     while ( not @{ $self->{_replies} } ) {
         my $ret = $self->{_socket}->recv( my $buffer, 131072 );
         unless ( defined $ret ) {
             next if $! == EINTR or $! == 0;
+            if ( $! == EINTR or $! == EWOULDBLOCK ) {
+                my $err = RedisDB::Error::EAGAIN->new("$!");
+                if ( $self->{raise_error} ) {
+                    die $err;
+                }
+                else {
+                    return $err;
+                }
+            }
             confess "Error reading reply from server: $!";
         }
         if ( $buffer ne '' ) {
@@ -455,7 +522,7 @@ sub get_reply {
     }
 
     my $res = shift @{ $self->{_replies} };
-    croak "$res" if ref $res eq 'RedisDB::Error';
+    croak "$res" if ref $res eq 'RedisDB::Error' and $self->{raise_error};
     return $res;
 }
 
@@ -553,6 +620,10 @@ my @commands = qw(
   zscore	zunionstore
 );
 
+sub _simple_commands {
+    return @commands;
+}
+
 =head1 WRAPPER METHODS
 
 Instead of using I<execute> and I<send_command> methods directly, it may be
@@ -665,16 +736,18 @@ not recommended.
 
 =head1 ERROR HANDLING
 
-If an error happens which the module can't handle, it will croak. It may
-happen as a result of a network error, or invalid data encoding, or if the
-server returned an error reply. In some cases the RedisDB object after throwing
-an exception will be left in inconsistant state. If you want to continue using
-the object after getting an exception, you should invoke the
-L</reset_connection> method on it. This will drop current connection and all
-outstanding requests, so the object will return to the same state it was just
-after creation with the L</new> method. If the connection was in subscription
-mode, you will have to restore all the subscriptions, if it was in the middle
-of transaction, you will have to start the transaction again.
+If an error happens which the module can't handle, it will throw an exception.
+It may happen as a result of a network error or invalid data encoding. Also
+module will throw an exception of the L<RedisDB::Error> type if the server
+returned an error reply and I<raise_error> parameter is set to true in the
+constructor (which is by default). After throwing an exception other than the
+L<RedisDB::Error> the RedisDB object will be left in inconsistent state. If you
+want to continue using the object after getting such an exception, you should
+invoke the L</reset_connection> method on it. This will drop current connection
+and all outstanding requests, so the object will return to the same state it
+was just after creation with the L</new> method. If the connection was in
+subscription mode, you will have to restore all the subscriptions, if it was in
+the middle of transaction, you will have to start the transaction again.
 
 =cut
 
@@ -684,12 +757,12 @@ Redis server may close a connection if it was idle for some time, also the
 connection may be closed in case when redis-server was restarted. RedisDB
 restores a connection to the server, but only if no data was lost as a result
 of the disconnect. E.g. if the client was idle for some time and the redis
-server closed the connection, it will be transparently restored on sending next
-command. If you sent a command and the server has closed the connection without
-sending a complete reply, the connection will not be restored and the module
-will throw an exception. Also the module will throw an exception if the
-connection was closed in the middle of a transaction or while you're in a
-subscription loop.
+server closed the connection, it will be transparently restored when you send a
+command next time.  If you sent a command and the server has closed the
+connection without sending a complete reply, the connection will not be
+restored and the module will throw an exception. Also the module will throw an
+exception if the connection was closed in the middle of a transaction or while
+you're in a subscription loop.
 
 =cut
 
@@ -709,7 +782,7 @@ will never be invoked:
     $redis->send_command( "ping", sub { $pong = $_[1] } );
     sleep 1 while not $pong;    # this will never return
 
-Therefore you need perriodically trigger check for the replies. The check is
+Therefore you need periodically trigger check for the replies. The check is
 triggered when you call the following methods: I<send_command>, I<reply_ready>,
 I<get_reply>, I<get_all_replies>.  Calling wrapper method, like
 C<< $redis->get('key') >>, will also trigger check as internally wrapper methods
@@ -800,7 +873,7 @@ You can publish messages into the channels using the I<publish> method. This
 method should be called when you in the normal mode, and can't be used while
 you're in the subscription mode.
 
-Following methods can be used in subscribtion mode:
+Following methods can be used in subscription mode:
 
 =cut
 
@@ -836,17 +909,17 @@ same as subscribe, but you specify patterns for channels' names.
 =back
 
 All parameters are optional, but you must subscribe at least to one channel. Also
-if default_callback is not specified, you have to explicitely specify a callback
+if default_callback is not specified, you have to explicitly specify a callback
 for every channel you are going to subscribe.
 
 =cut
 
 sub subscription_loop {
     my ( $self, %args ) = @_;
-    croak "Already in subscription loop" if $self->{_subscribtion_loop};
+    croak "Already in subscription loop" if $self->{_subscription_loop};
     croak "You can't start subscription loop while in pipelining mode."
       if $self->{_parser}->callbacks
-          or @{ $self->{_replies} };
+      or @{ $self->{_replies} };
     $self->{_subscribed}        = {};
     $self->{_psubscribed}       = {};
     $self->{_subscription_cb}   = $args{default_callback};
@@ -1025,14 +1098,14 @@ sub psubscribed {
 Transactions allow you to execute a sequence of commands in a single step. In
 order to start a transaction you should use the I<multi> method.  After you
 have entered a transaction all the commands you issue are queued, but not
-executed till you call the I<exec> method. Tipically these commands return
-string "QUEUED" as result, but if there is an error in e.g. number of
+executed till you call the I<exec> method. Typically these commands return
+string "QUEUED" as a result, but if there is an error in e.g. number of
 arguments, they may croak. When you call exec, all the queued commands will be
 executed and exec will return a list of results for every command in the
-transaction. If any command has failed, exec will croak. If instead of I<exec>
-you call I<discard>, all scheduled commands will be canceled.
+transaction. If instead of I<exec> you call I<discard>, all scheduled commands
+will be canceled.
 
-You can set some keys as watched. If any whatched key has been changed by
+You can set some keys as watched. If any watched key has been changed by
 another client before you called exec, the transaction will be discarded and
 exec will return false value.
 
@@ -1090,7 +1163,7 @@ __END__
 
 =head1 SEE ALSO
 
-L<Redis>, L<Redis::hiredis>, L<AnyEvent::Redis>
+L<Redis>, L<Redis::hiredis>, L<Redis::Client>, L<AnyEvent::Redis>, L<AnyEvent::Redis::RipeRedis>
 
 =head1 WHY ANOTHER ONE
 
@@ -1098,9 +1171,9 @@ I was in need of the client for redis database. L<AnyEvent::Redis> didn't suite
 me as it requires an event loop, and it didn't fit into the existing code. The
 problem with L<Redis> is that it didn't (at the time I started this) reconnect
 to the server if connection was closed after timeout or as result of the server
-restart, and it does not support pipelining. After analizing what I need to
+restart, and it does not support pipelining. After analyzing what I need to
 change in L<Redis> in order to get all I want, I decided that it will be
-simplier to write the new module from scratch. This also solves the problem of
+simpler to write the new module from scratch. This also solves the problem of
 backward compatibility.
 
 =head1 BUGS
